#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "error.h"

#define MAXLINE   512
#define MAXMESG   2048


/*
** Read n bytes from a descriptor.
** Use in place of read() when fd is a stream socket.
*/
int readn(register int fd, register char *ptr, register int nbytes)
{
  int nleft, nread;
  
  nleft = nbytes;
  while(nleft > 0)
  {
    if( (nread = read(fd, ptr, nleft)) < 0)
      return(nread);
    else if(nread == 0)
      break;  /* EOF */
      
    nleft -= nread;
    ptr += nread;
  }
  return(nbytes - nleft);
}


/*
** Write n bytes to a desc..
** Use in place of write() when fd is a stream socket.
*/
int writen(register int fd, register char *ptr, register int nbytes)
{
  int nleft, nwritten;

  nleft = nbytes;
  while(nleft > 0)
  {
    if( (nwritten = write(fd, ptr, nleft)) <= 0)
      return(nwritten);
                        
    nleft -= nwritten;
    ptr += nwritten;
  }
  return(nbytes - nleft);
}


/*
** Read a line from a desc.. Read the line one byte at a time, looking for the
** newline. We store the newline in the buffer, then follow it with a null
** (the same as fgets(3)).
** We return the number of chars up to, but not including, the null (the same
** as strlen(3)).
*/
int readline(register int fd, register char *ptr, register int maxlen)
{
  int   n, rc;
  char  c;
  
  for(n = 1; n < maxlen; n++)
  {
    if( (rc = read(fd, &c, 1)) == 1)
    {
      *ptr++ = c;
      if(c == '\n')
        break;
    }
    else if(rc == 0)
    {
      if(n == 1)
        return(0);    /* EOF, no data read */
      else
        break;        /* EOF, some data was read */
    }
    else
      return(-1);     /* error */
  }
  
  *ptr = 0;
  return(n);
}


/*
** Read a stream socket one line at a time, and write each line back to
** the sender.
**
** Return when the connection is terminated.
*/
void str_echo(int sockfd)
{
  int   n;
  char  line[MAXLINE];
  
  while(1)
  {
    if( (n = readline(sockfd, line, MAXLINE)) < 0)
      err_mesg(DUMP, "str_echo: readline error\n");
    else if(n == 0)
      return;       /* connection closed */
      
    if(writen(sockfd, line, n) != n)
      err_mesg(DUMP, "str_echo: writen error\n");
  }
}


/*
** Read the contens of the FILE *fp, write each line to the stream socket
** (to the server process), then read a line back from the socket and write
** it to stdout.
*/
void str_cli(register FILE *fp, register int sockfd)
{
  int   n;
  char  sendline[MAXLINE], recvline[MAXLINE+1];
  
  while(fgets(sendline, MAXLINE, fp) != NULL)
  {
    n = strlen(sendline);
    if(writen(sockfd, sendline, n) != n)
      err_mesg(FATAL_SYS, "str_cli: writen error.\n");
      
    /*
    ** Now read a line from the socket and write it ti our stdout
    */
    
    if( (n = readline(sockfd, recvline, MAXLINE)) < 0)
      err_mesg(DUMP, "str_cli: readline error\n");
    recvline[n] = 0;
    fputs(recvline, stdout);
  }
  if(ferror(fp))
    err_mesg(FATAL_SYS, "str_cli: error reading file\n");
}


/*
** Read a datagram from a connectionless socket and write it back to the
** sender.
**
** We never return, as we never know when a datagram client is done.
*/
void dg_echo(int sockfd, struct sockaddr *pcli_addr, int maxclilen)
{
  int   n, clilen;
  char  mesg[MAXMESG];
  
  while(1)
  {
    clilen = maxclilen;
    if( (n = recvfrom(sockfd, mesg, MAXMESG, 0, pcli_addr, &clilen)) < 0)
      err_mesg(DUMP, "dg_echo: recvfrom error\n");
      
    if(sendto(sockfd, mesg, n, 0, pcli_addr, clilen) != n)
      err_mesg(DUMP, "dg_echo: sendto error\n");
  }
}


/*
** Read the contens of the FILE *fp, write each line to the datagram socket,
** then read a line back from the socket and write
** it to stdout.
**
** Return to caller when an EOF is encountered on the input file.
*/
void dg_cli(FILE *fp, int sockfd, struct sockaddr *pserv_addr, int servlen)
{
  int   n;
  char  sendline[MAXLINE], recvline[MAXLINE+1];

  while(fgets(sendline, MAXLINE, fp) != NULL)
    {
    n = strlen(sendline);
    if(sendto(sockfd, sendline, n, 0, pserv_addr, servlen) != n)
      err_mesg(FATAL_SYS, "dg_cli: sendto error.\n");

    /*
    ** Now read a line from the socket and write it ti our stdout
    */

    if( (n = recvfrom(sockfd, recvline, MAXLINE, 0, NULL, NULL)) < 0)
      err_mesg(DUMP, "dg_cli: recvfrom error\n");
    recvline[n] = 0;
    fputs(recvline, stdout);
  }
  if(ferror(fp))
    err_mesg(FATAL_SYS, "dg_cli: error reading file\n");
}
